# Copyright 2017-present Kensho Technologies, LLC.
"""Front-end for GraphQL to database queries compiler.

High-level overview of the GraphQL ingestion process that outputs the compiler's
internal representation (IR) via the graphql_to_ir() function:
    - The function receives a GraphQL string and a GraphQL schema.
      How the schema is constructed is beyond the scope of this package.
    - It uses the graphql library to parse the GraphQL string into
      an abstract syntax tree representation (AST).
    - It validates the GraphQL string using the schema.
    - Finally, it converts the validated GraphQL string into an internal representation (IR):
      a list of BasicBlock (blocks.py) objects, which may further contain
      Expression (expressions.py) objects.

To get from GraphQL AST to IR, we follow the following pattern:
    step 0. preprocessing:  (see _compile_ast_node_to_ir())
        - read any directives at the current AST node;
        - read the inline fragment at the current AST node, if one exists;
        - read any child fields of the current AST node, splitting them into two groups:
          property fields and vertex fields; all property fields must precede the vertex fields
          for the AST to be valid;

    step 1. apply all @filter directives that apply to the current field
            (see _compile_ast_node_to_ir() and directive_helpers.get_local_filter_directives())

    We now proceed with one of three cases (P, V and F), depending on whether
    the current AST node is a property AST node, vertex AST node, or inline fragment, respectively.
    The root AST node is always a vertex AST node.

    *** P-steps ***
    step P-2. Process property-only directives, like @tag and @output.
    ***************

    *** V-steps ***
    step V-2. Recurse into any property field children of the current AST node
              (property fields cannot have property fields of their own, see _compile_vertex_ast()).

    step V-3. Property field processing complete:  (see _compile_vertex_ast())
        - mark the current location in the query, since all @filter directives that apply to the
          current field have already been processed;
        - process the output_source directive, if it exists

    step V-4. Recurse into any vertex field children of the current AST node:
              (see _compile_vertex_ast())
        - before recursing into each vertex:
          - process any @optional and @fold directives present on the child AST node;
          - process any @output within a @fold context,
            and prevent further traversal if one is present;
        - after returning from each vertex:
          - return to the marked query location using the appropriate manner,
            depending on whether @optional was present on the child or not;
          - if the visited vertex had a @fold or @optional directive,
            undo the processing performed at previous steps;
    ***************

    *** F-steps ***
    step F-2. Emit a type coercion block if appropriate, then recurse into the fragment's selection.
    ***************
"""
from collections import namedtuple

from graphql import (
    GraphQLInt, GraphQLInterfaceType, GraphQLList, GraphQLObjectType, GraphQLUnionType
)
from graphql.error import GraphQLSyntaxError
from graphql.language.ast import Field, InlineFragment
from graphql.language.parser import parse
from graphql.validation import validate
import six

from . import blocks, expressions
from ..exceptions import GraphQLCompilationError, GraphQLParsingError, GraphQLValidationError
from ..schema import COUNT_META_FIELD_NAME, DIRECTIVES
from .context_helpers import (
    get_context_fold_info, get_optional_scope_or_none, has_encountered_output_source,
    has_fold_count_filter, is_in_fold_innermost_scope, is_in_fold_scope, is_in_optional_scope,
    set_fold_count_filter, set_fold_innermost_scope, set_fold_scope_data, set_optional_scope_data,
    set_output_source_data, unmark_context_fold_scope, unmark_fold_count_filter,
    unmark_fold_innermost_scope, unmark_optional_scope, validate_context_for_visiting_vertex_field
)
from .directive_helpers import (
    get_local_filter_directives, get_unique_directives, validate_property_directives,
    validate_root_vertex_directives, validate_vertex_directives,
    validate_vertex_field_directive_in_context, validate_vertex_field_directive_interactions
)
from .filters import process_filter_directive
from .helpers import (
    FoldScopeLocation, Location, get_ast_field_name, get_edge_direction_and_name,
    get_field_type_from_schema, get_uniquely_named_objects_by_name, get_vertex_field_type,
    invert_dict, is_vertex_field_name, strip_non_null_from_type, validate_output_name,
    validate_safe_string
)
from .metadata import LocationInfo, QueryMetadataTable, RecurseInfo


# LocationStackEntry contains the following:
# - location: Location object corresponding to an inserted MarkLocation block
# - num_traverses: Int counter for the number of traverses inserted after the last MarkLocation
#                  (corresponding Location stored in `location`)
LocationStackEntry = namedtuple('LocationStackEntry', ('location', 'num_traverses'))


def _construct_location_stack_entry(location, num_traverses):
    """Return a LocationStackEntry namedtuple with the specified parameters."""
    if not isinstance(num_traverses, int) or num_traverses < 0:
        raise AssertionError(u'Attempted to create a LocationStackEntry namedtuple with an invalid '
                             u'value for "num_traverses" {}. This is not allowed.'
                             .format(num_traverses))
    if not isinstance(location, Location):
        raise AssertionError(u'Attempted to create a LocationStackEntry namedtuple with an invalid '
                             u'value for "location" {}. This is not allowed.'
                             .format(location))
    return LocationStackEntry(location=location, num_traverses=num_traverses)


# The OutputMetadata will have the following types for its members:
# - type: a GraphQL type object, like String or Integer, describing the type of that output value
# - optional: boolean, whether the output is part of an optional traversal and
#             could therefore have a value of null because it did not exist
class OutputMetadata(namedtuple('OutputMetadata', ('type', 'optional'))):
    def __eq__(self, other):
        """Check another OutputMetadata object for equality against this one."""
        # Unfortunately, GraphQL types don't have an equality operator defined,
        # and instead have this "is_same_type" function. Hence, we have to override equality here.
        return self.type.is_same_type(other.type) and self.optional == other.optional

    def __ne__(self, other):
        """Check another OutputMetadata object for non-equality against this one."""
        return not self.__eq__(other)


IrAndMetadata = namedtuple(
    'IrAndMetadata', (
        'ir_blocks',
        'input_metadata',
        'output_metadata',
        'query_metadata_table',
    )
)


def _get_fields(ast):
    """Return a list of vertex fields, and a list of property fields, for the given AST node.

    Also verifies that all property fields for the AST node appear before all vertex fields,
    raising GraphQLCompilationError if that is not the case.

    Args:
        ast: GraphQL AST node, obtained from the graphql library

    Returns:
        tuple of two lists
            - the first list contains ASTs for vertex fields
            - the second list contains ASTs for property fields
    """
    if not ast.selection_set:
        # There are no child fields.
        return [], []

    property_fields = []
    vertex_fields = []
    seen_field_names = set()
    switched_to_vertices = False  # Ensures that all property fields are before all vertex fields.
    for field_ast in ast.selection_set.selections:
        if not isinstance(field_ast, Field):
            # We are getting Fields only, ignore everything else.
            continue

        name = get_ast_field_name(field_ast)
        if name in seen_field_names:
            # If we ever allow repeated field names,
            # then we have to change the Location naming scheme to reflect the repetitions
            # and disambiguate between Recurse and Traverse visits to a Location.
            raise GraphQLCompilationError(u'Encountered repeated field name: {}'.format(name))
        seen_field_names.add(name)

        # Vertex fields start with 'out_' or 'in_', denoting the edge direction to that vertex.
        if is_vertex_field_name(name):
            switched_to_vertices = True
            vertex_fields.append(field_ast)
        else:
            if switched_to_vertices:
                raise GraphQLCompilationError(u'Encountered property field {} '
                                              u'after vertex fields!'.format(name))
            property_fields.append(field_ast)

    return vertex_fields, property_fields


def _get_inline_fragment(ast):
    """Return the inline fragment at the current AST node, or None if no fragment exists."""
    if not ast.selection_set:
        # There is nothing selected here, so no fragment.
        return None

    fragments = [
        ast_node
        for ast_node in ast.selection_set.selections
        if isinstance(ast_node, InlineFragment)
    ]

    if not fragments:
        return None

    if len(fragments) > 1:
        raise GraphQLCompilationError(u'Cannot compile GraphQL with more than one fragment in '
                                      u'a given selection set.')

    return fragments[0]


def _mark_location(location):
    """Return a MarkLocation basic block that marks the present location in the query."""
    return blocks.MarkLocation(location)


def _process_output_source_directive(schema, current_schema_type, ast,
                                     location, context, local_unique_directives):
    """Process the output_source directive, modifying the context as appropriate.

    Args:
        schema: GraphQL schema object, obtained from the graphql library
        current_schema_type: GraphQLType, the schema type at the current location
        ast: GraphQL AST node, obtained from the graphql library
        location: Location object representing the current location in the query
        context: dict, various per-compilation data (e.g. declared tags, whether the current block
                 is optional, etc.). May be mutated in-place in this function!
        local_unique_directives: dict, directive name string -> directive object, containing
                                 unique directives present on the current AST node *only*

    Returns:
        an OutputSource block, if one should be emitted, or None otherwise
    """
    # The 'ast' variable is only for function signature uniformity, and is currently not used.
    output_source_directive = local_unique_directives.get('output_source', None)
    if output_source_directive:
        if has_encountered_output_source(context):
            raise GraphQLCompilationError(u'Cannot have more than one output source!')
        if is_in_optional_scope(context):
            raise GraphQLCompilationError(u'Cannot have the output source in an optional block!')
        set_output_source_data(context, location)
        return blocks.OutputSource()
    else:
        return None


def _compile_property_ast(schema, current_schema_type, ast, location,
                          context, unique_local_directives):
    """Process property directives at this AST node, updating the query context as appropriate.

    Args:
        schema: GraphQL schema object, obtained from the graphql library
        current_schema_type: GraphQLType, the schema type at the current location
        ast: GraphQL AST node, obtained from the graphql library. Only for function signature
             uniformity at the moment -- it is currently not used.
        location: Location object representing the current location in the query
        context: dict, various per-compilation data (e.g. declared tags, whether the current block
                 is optional, etc.). May be mutated in-place in this function!
        unique_local_directives: dict, directive name string -> directive object, containing
                                 unique directives present on the current AST node *only*
    """
    validate_property_directives(unique_local_directives)

    if location.field == COUNT_META_FIELD_NAME:
        # Verify that uses of this field are within a @fold scope.
        if not is_in_fold_scope(context):
            raise GraphQLCompilationError(u'Cannot use the "{}" meta field when not within a @fold '
                                          u'vertex field, as counting elements only makes sense '
                                          u'in a fold. Location: {}'
                                          .format(COUNT_META_FIELD_NAME, location))

    # step P-2: process property-only directives
    tag_directive = unique_local_directives.get('tag', None)
    if tag_directive:
        if is_in_fold_scope(context):
            raise GraphQLCompilationError(u'Tagging values within a @fold vertex field is '
                                          u'not allowed! Location: {}'.format(location))

        if location.field == COUNT_META_FIELD_NAME:
            raise AssertionError(u'Tags are prohibited within @fold, but unexpectedly found use of '
                                 u'a tag on the {} meta field that is only allowed within a @fold!'
                                 u'Location: {}'
                                 .format(COUNT_META_FIELD_NAME, location))

        # Schema validation has ensured that the fields below exist.
        tag_name = tag_directive.arguments[0].value.value
        if tag_name in context['tags']:
            raise GraphQLCompilationError(u'Cannot reuse tag name: {}'.format(tag_name))
        validate_safe_string(tag_name)
        context['tags'][tag_name] = {
            'location': location,
            'optional': is_in_optional_scope(context),
            'type': strip_non_null_from_type(current_schema_type),
        }

    output_directive = unique_local_directives.get('output', None)
    if output_directive:
        # Schema validation has ensured that the fields below exist.
        output_name = output_directive.arguments[0].value.value
        if output_name in context['outputs']:
            raise GraphQLCompilationError(u'Cannot reuse output name: '
                                          u'{}, {}'.format(output_name, context))
        validate_safe_string(output_name)
        validate_output_name(output_name)

        graphql_type = strip_non_null_from_type(current_schema_type)
        if is_in_fold_scope(context):
            # Fold outputs are only allowed at the last level of traversal.
            set_fold_innermost_scope(context)

            if location.field != COUNT_META_FIELD_NAME:
                graphql_type = GraphQLList(graphql_type)

        context['outputs'][output_name] = {
            'location': location,
            'optional': is_in_optional_scope(context),
            'type': graphql_type,
            'fold': context.get('fold', None),
        }


def _get_recurse_directive_depth(field_name, field_directives):
    """Validate and return the depth parameter of the recurse directive."""
    recurse_directive = field_directives['recurse']
    optional_directive = field_directives.get('optional', None)

    if optional_directive:
        raise GraphQLCompilationError(u'Found both @optional and @recurse on '
                                      u'the same vertex field: {}'.format(field_name))

    recurse_args = get_uniquely_named_objects_by_name(recurse_directive.arguments)
    recurse_depth = int(recurse_args['depth'].value.value)
    if recurse_depth < 1:
        raise GraphQLCompilationError(u'Found recurse directive with disallowed depth: '
                                      u'{}'.format(recurse_depth))

    return recurse_depth


def _validate_recurse_directive_types(current_schema_type, field_schema_type, context):
    """Perform type checks on the enclosing type and the recursed type for a recurse directive.

    Args:
        current_schema_type: GraphQLType, the schema type at the current location
        field_schema_type: GraphQLType, the schema type at the inner scope
        context: dict, various per-compilation data (e.g. declared tags, whether the current block
                 is optional, etc.). May be mutated in-place in this function!
    """
    # Get the set of all allowed types in the current scope.
    type_hints = context['type_equivalence_hints'].get(field_schema_type)
    type_hints_inverse = context['type_equivalence_hints_inverse'].get(field_schema_type)
    allowed_current_types = {field_schema_type}

    if type_hints and isinstance(type_hints, GraphQLUnionType):
        allowed_current_types.update(type_hints.types)

    if type_hints_inverse and isinstance(type_hints_inverse, GraphQLUnionType):
        allowed_current_types.update(type_hints_inverse.types)

    # The current scope must be of the same type as the field scope, or an acceptable subtype.
    current_scope_is_allowed = current_schema_type in allowed_current_types

    is_implemented_interface = (
        isinstance(field_schema_type, GraphQLInterfaceType) and
        isinstance(current_schema_type, GraphQLObjectType) and
        field_schema_type in current_schema_type.interfaces
    )

    if not any((current_scope_is_allowed, is_implemented_interface)):
        raise GraphQLCompilationError(u'Edges expanded with a @recurse directive must either '
                                      u'be of the same type as their enclosing scope, a supertype '
                                      u'of the enclosing scope, or be of an interface type that is '
                                      u'implemented by the type of their enclosing scope. '
                                      u'Enclosing scope type: {}, edge type: '
                                      u'{}'.format(current_schema_type, field_schema_type))


def _compile_vertex_ast(schema, current_schema_type, ast,
                        location, context, unique_local_directives, fields):
    """Return a list of basic blocks corresponding to the vertex AST node.

    Args:
        schema: GraphQL schema object, obtained from the graphql library
        current_schema_type: GraphQLType, the schema type at the current location
        ast: GraphQL AST node, obtained from the graphql library
        location: Location object representing the current location in the query
        context: dict, various per-compilation data (e.g. declared tags, whether the current block
                 is optional, etc.). May be mutated in-place in this function!
        unique_local_directives: dict, directive name string -> directive object, containing
                                 unique directives present on the current AST node *only*
        fields: tuple of lists (property_fields, vertex_fields), with lists of field objects
                present on the current vertex AST node

    Returns:
        list of basic blocks, the compiled output of the vertex AST node
    """
    basic_blocks = []
    query_metadata_table = context['metadata']
    current_location_info = query_metadata_table.get_location_info(location)

    vertex_fields, property_fields = fields

    validate_vertex_directives(unique_local_directives)

    # step V-2: step into property fields
    for field_ast in property_fields:
        field_name = get_ast_field_name(field_ast)
        property_schema_type = get_field_type_from_schema(current_schema_type, field_name)

        inner_location = location.navigate_to_field(field_name)
        inner_basic_blocks = _compile_ast_node_to_ir(schema, property_schema_type, field_ast,
                                                     inner_location, context)
        basic_blocks.extend(inner_basic_blocks)

    # step V-3: mark the graph position, and process output_source directive
    basic_blocks.append(_mark_location(location))

    output_source = _process_output_source_directive(schema, current_schema_type, ast,
                                                     location, context, unique_local_directives)
    if output_source:
        basic_blocks.append(output_source)

    # step V-4: step into vertex fields
    for field_ast in vertex_fields:
        field_name = get_ast_field_name(field_ast)
        validate_context_for_visiting_vertex_field(location, field_name, context)

        field_schema_type = get_vertex_field_type(current_schema_type, field_name)
        hinted_base = context['type_equivalence_hints_inverse'].get(field_schema_type, None)
        if hinted_base:
            field_schema_type = hinted_base

        inner_unique_directives = get_unique_directives(field_ast)
        validate_vertex_field_directive_interactions(location, field_name, inner_unique_directives)
        validate_vertex_field_directive_in_context(
            location, field_name, inner_unique_directives, context)

        recurse_directive = inner_unique_directives.get('recurse', None)
        optional_directive = inner_unique_directives.get('optional', None)
        fold_directive = inner_unique_directives.get('fold', None)
        in_topmost_optional_block = False

        edge_traversal_is_optional = optional_directive is not None
        edge_traversal_is_folded = fold_directive is not None
        edge_traversal_is_recursive = recurse_directive is not None

        # This is true for any vertex expanded within an @optional scope.
        within_optional_scope = is_in_optional_scope(context)

        if edge_traversal_is_optional:
            # Invariant: There must always be a marked location corresponding to the query position
            # immediately before any optional Traverse.
            #
            # This invariant is verified in the IR sanity checks module (ir_sanity_checks.py),
            # in the function named _sanity_check_mark_location_preceding_optional_traverse().
            #
            # This marked location is the one that the @optional directive's corresponding
            # optional Backtrack will jump back to. If such a marked location isn't present,
            # the backtrack could rewind to an old marked location and might ignore
            # entire stretches of applied filtering.
            #
            # Assumption: The only way there might not be a marked location here is
            # if the current location already traversed into child locations, not including folds.
            non_fold_child_locations = {
                child_location
                for child_location in query_metadata_table.get_child_locations(location)
                if not isinstance(child_location, FoldScopeLocation)
            }
            if non_fold_child_locations:
                location = query_metadata_table.revisit_location(location)
                basic_blocks.append(_mark_location(location))

        if fold_directive:
            inner_location = location.navigate_to_fold(field_name)
        else:
            inner_location = location.navigate_to_subpath(field_name)

        inner_location_info = LocationInfo(
            parent_location=location,
            type=strip_non_null_from_type(field_schema_type),
            coerced_from_type=None,
            optional_scopes_depth=(
                current_location_info.optional_scopes_depth + edge_traversal_is_optional),
            recursive_scopes_depth=(
                current_location_info.recursive_scopes_depth + edge_traversal_is_recursive),
            is_within_fold=(current_location_info.is_within_fold or edge_traversal_is_folded),
        )
        query_metadata_table.register_location(inner_location, inner_location_info)

        if edge_traversal_is_optional:
            # Remember where the topmost optional context started.
            topmost_optional = get_optional_scope_or_none(context)
            if topmost_optional is None:
                set_optional_scope_data(context, inner_location)
                in_topmost_optional_block = True

        edge_direction, edge_name = get_edge_direction_and_name(field_name)

        if fold_directive:
            fold_block = blocks.Fold(inner_location)
            basic_blocks.append(fold_block)
            set_fold_scope_data(context, inner_location)
        elif recurse_directive:
            _validate_recurse_directive_types(current_schema_type, field_schema_type, context)
            recurse_depth = _get_recurse_directive_depth(field_name, inner_unique_directives)
            basic_blocks.append(blocks.Recurse(edge_direction,
                                               edge_name,
                                               recurse_depth,
                                               within_optional_scope=within_optional_scope))
            query_metadata_table.record_recurse_info(location,
                                                     RecurseInfo(edge_direction=edge_direction,
                                                                 edge_name=edge_name,
                                                                 depth=recurse_depth))
        else:
            basic_blocks.append(blocks.Traverse(edge_direction, edge_name,
                                                optional=edge_traversal_is_optional,
                                                within_optional_scope=within_optional_scope))

        inner_basic_blocks = _compile_ast_node_to_ir(schema, field_schema_type, field_ast,
                                                     inner_location, context)
        basic_blocks.extend(inner_basic_blocks)

        if edge_traversal_is_folded:
            has_count_filter = has_fold_count_filter(context)
            _validate_fold_has_outputs_or_count_filter(
                get_context_fold_info(context), has_count_filter, context['outputs'])
            basic_blocks.append(blocks.Unfold())
            unmark_context_fold_scope(context)
            if has_count_filter:
                unmark_fold_count_filter(context)
            if is_in_fold_innermost_scope(context):
                unmark_fold_innermost_scope(context)

        if in_topmost_optional_block:
            unmark_optional_scope(context)

        # If we are currently evaluating a @fold vertex,
        # we didn't Traverse into it, so we don't need to backtrack out either.
        # We also don't backtrack if we've reached an @output_source.
        backtracking_required = (
            (not fold_directive) and
            (not has_encountered_output_source(context))
        )
        if backtracking_required:
            if edge_traversal_is_optional:
                basic_blocks.append(blocks.EndOptional())
                basic_blocks.append(blocks.Backtrack(location, optional=True))

                # Exiting optional block!
                # Revisit the location so that there is a marked location right after the optional,
                # so that future Backtrack blocks return after the optional set of blocks, and
                # don't accidentally return to a prior location instead.
                location = query_metadata_table.revisit_location(location)

                basic_blocks.append(_mark_location(location))
            else:
                basic_blocks.append(blocks.Backtrack(location))

    return basic_blocks


def _validate_fold_has_outputs_or_count_filter(fold_scope_location, fold_has_count_filter, outputs):
    """Ensure the @fold scope has at least one output, or filters on the size of the fold."""
    # This function makes sure that the @fold scope has an effect.
    # Folds either output data, or filter the data enclosing the fold based on the size of the fold.
    if fold_has_count_filter:
        # This fold has a filter on the "_x_count" property, so it is legal and has an effect.
        return True

    # At least one output in the outputs list must point to the fold_scope_location,
    # or the scope corresponding to fold_scope_location had no @outputs and is illegal.
    for output in six.itervalues(outputs):
        if output['fold'] == fold_scope_location:
            return True

    raise GraphQLCompilationError(u'Found a @fold scope that has no effect on the query. '
                                  u'Each @fold scope must either perform filtering, or contain at '
                                  u'least one field marked for output. Fold location: {}'
                                  .format(fold_scope_location))


def _compile_fragment_ast(schema, current_schema_type, ast, location, context):
    """Return a list of basic blocks corresponding to the inline fragment at this AST node.

    Args:
        schema: GraphQL schema object, obtained from the graphql library
        current_schema_type: GraphQLType, the schema type at the current location
        ast: GraphQL AST node, obtained from the graphql library.
        location: Location object representing the current location in the query
        context: dict, various per-compilation data (e.g. declared tags, whether the current block
                 is optional, etc.). May be mutated in-place in this function!

    Returns:
        list of basic blocks, the compiled output of the vertex AST node
    """
    query_metadata_table = context['metadata']

    # step F-2. Emit a type coercion block if appropriate,
    #           then recurse into the fragment's selection.
    coerces_to_type_name = ast.type_condition.name.value
    coerces_to_type_obj = schema.get_type(coerces_to_type_name)

    basic_blocks = []

    # Check if the coercion is necessary.
    # No coercion is necessary if coercing to the current type of the scope,
    # or if the scope is of union type, to the base type of the union as defined by
    # the type_equivalence_hints compilation parameter.
    is_same_type_as_scope = current_schema_type.is_same_type(coerces_to_type_obj)
    equivalent_union_type = context['type_equivalence_hints'].get(coerces_to_type_obj, None)
    is_base_type_of_union = (
        isinstance(current_schema_type, GraphQLUnionType) and
        current_schema_type.is_same_type(equivalent_union_type)
    )

    if not (is_same_type_as_scope or is_base_type_of_union):
        # Coercion is required.
        query_metadata_table.record_coercion_at_location(location, coerces_to_type_obj)
        basic_blocks.append(blocks.CoerceType({coerces_to_type_name}))

    inner_basic_blocks = _compile_ast_node_to_ir(
        schema, coerces_to_type_obj, ast, location, context)
    basic_blocks.extend(inner_basic_blocks)

    return basic_blocks


def _compile_ast_node_to_ir(schema, current_schema_type, ast, location, context):
    """Compile the given GraphQL AST node into a list of basic blocks.

    Args:
        schema: GraphQL schema object, obtained from the graphql library
        current_schema_type: GraphQLType, the schema type at the current location
        ast: the current GraphQL AST node, obtained from the graphql library
        location: Location object representing the current location in the query
        context: dict, various per-compilation data (e.g. declared tags, whether the current block
                 is optional, etc.). May be mutated in-place in this function!

    Returns:
        list of basic blocks corresponding to this GraphQL AST node
    """
    basic_blocks = []

    # step 0: preprocessing
    local_unique_directives = get_unique_directives(ast)
    fields = _get_fields(ast)
    vertex_fields, property_fields = fields
    fragment = _get_inline_fragment(ast)
    filter_operations = get_local_filter_directives(
        ast, current_schema_type, vertex_fields)

    # We don't support type coercion while at the same time selecting fields.
    # Either there are no fields, or there is no fragment, otherwise we raise a compilation error.
    fragment_exists = fragment is not None
    fields_exist = vertex_fields or property_fields
    if fragment_exists and fields_exist:
        raise GraphQLCompilationError(u'Cannot compile GraphQL that has inline fragment and '
                                      u'selected fields in the same selection. Please move the '
                                      u'selected fields inside the inline fragment.')

    if location.field is not None:  # we're at a property field
        # sanity-check: cannot have an inline fragment at a property field
        if fragment_exists:
            raise AssertionError(u'Found inline fragment at a property field: '
                                 u'{} {}'.format(location, fragment))

        # sanity-check: locations at properties don't have their own property locations
        if len(property_fields) > 0:
            raise AssertionError(u'Found property fields on a property field: '
                                 u'{} {}'.format(location, property_fields))

    # step 1: apply local filter, if any
    for filter_operation_info in filter_operations:
        filter_block = process_filter_directive(filter_operation_info, location, context)
        if isinstance(location, FoldScopeLocation) and location.field == COUNT_META_FIELD_NAME:
            # Filtering on the fold count field is only allowed at the innermost scope of a fold.
            set_fold_innermost_scope(context)

            # This Filter is going in the global operations section of the query, so it cannot
            # use LocalField expressions since there is no "local" location to use.
            # Rewrite it so that all references of data at a location instead use ContextFields.
            expected_field = expressions.LocalField(COUNT_META_FIELD_NAME)
            replacement_field = expressions.FoldedContextField(location, GraphQLInt)

            visitor_fn = expressions.make_replacement_visitor(expected_field, replacement_field)
            filter_block = filter_block.visit_and_update_expressions(visitor_fn)

            visitor_fn = expressions.make_type_replacement_visitor(
                expressions.ContextField,
                lambda context_field: expressions.GlobalContextField(context_field.location))
            filter_block = filter_block.visit_and_update_expressions(visitor_fn)

            set_fold_count_filter(context)
            context['global_filters'].append(filter_block)
        else:
            basic_blocks.append(filter_block)

    if location.field is not None:
        # The location is at a property, compile the property data following P-steps.
        _compile_property_ast(schema, current_schema_type, ast,
                              location, context, local_unique_directives)
    else:
        # The location is at a vertex.
        if fragment_exists:
            # Compile the fragment data following F-steps.
            # N.B.: Note that the "fragment" variable is the fragment's AST. Since we've asserted
            #       that the fragment is the only part of the selection set at the current AST node,
            #       we pass the "fragment" in the AST parameter of the _compile_fragment_ast()
            #       function, rather than the current AST node as in the other compilation steps.
            basic_blocks.extend(
                _compile_fragment_ast(schema, current_schema_type, fragment, location, context))
        else:
            # Compile the vertex data following V-steps.
            basic_blocks.extend(
                _compile_vertex_ast(schema, current_schema_type, ast,
                                    location, context, local_unique_directives, fields))

    return basic_blocks


def _compile_root_ast_to_ir(schema, ast, type_equivalence_hints=None):
    """Compile a full GraphQL abstract syntax tree (AST) to intermediate representation.

    Args:
        schema: GraphQL schema object, obtained from the graphql library
        ast: the root GraphQL AST node for the query, obtained from the graphql library,
             and already validated against the schema for type-correctness
        type_equivalence_hints: optional dict of GraphQL type to equivalent GraphQL union

    Returns:
        IrAndMetadata named tuple, containing fields:
        - ir_blocks: a list of IR basic block objects
        - input_metadata: a dict of expected input parameters (string) -> inferred GraphQL type
        - output_metadata: a dict of output name (string) -> OutputMetadata object
        - location_types: a dict of location objects -> GraphQL type objects at that location
        - coerced_locations: a set of location objects indicating where type coercions have happened
    """
    if len(ast.selection_set.selections) != 1:
        raise GraphQLCompilationError(u'Cannot process AST with more than one root selection!')

    base_ast = ast.selection_set.selections[0]
    base_start_type = get_ast_field_name(base_ast)  # This is the type at which querying starts.

    # Validation passed, so the base_start_type must exist as a field of the root query.
    current_schema_type = get_field_type_from_schema(schema.get_query_type(), base_start_type)

    # Construct the start location of the query and its associated metadata.
    location = Location((base_start_type,))
    base_location_info = LocationInfo(
        parent_location=None,
        type=current_schema_type,
        coerced_from_type=None,
        optional_scopes_depth=0,
        recursive_scopes_depth=0,
        is_within_fold=False,
    )
    query_metadata_table = QueryMetadataTable(location, base_location_info)

    # Default argument value is empty dict
    if not type_equivalence_hints:
        type_equivalence_hints = dict()

    # Construct the starting context object.
    context = {
        # 'metadata' is the QueryMetadataTable describing all the metadata collected during query
        # processing, including location metadata (e.g. which locations are folded or optional).
        'metadata': query_metadata_table,
        # 'tags' is a dict containing
        #  - location: Location where the tag was defined
        #  - optional: boolean representing whether the tag was defined within an @optional scope
        #  - type: GraphQLType of the tagged value
        'tags': dict(),
        # 'global_filters' is a list that may contain Filter blocks that are generated during
        # query processing, but apply to the global query scope and should be appended to the
        # IR blocks only after the GlobalOperationsStart block has been emitted.
        'global_filters': [],
        # 'outputs' is a dict mapping each output name to another dict which contains
        #  - location: Location where to output from
        #  - optional: boolean representing whether the output was defined within an @optional scope
        #  - type: GraphQLType of the output
        #  - fold: FoldScopeLocation object if the current output was defined within a fold scope,
        #          and None otherwise
        'outputs': dict(),
        # 'inputs' is a dict mapping input parameter names to their respective expected GraphQL
        # types, as automatically inferred by inspecting the query structure
        'inputs': dict(),
        # 'type_equivalence_hints' is a dict mapping GraphQL types to equivalent GraphQL unions
        'type_equivalence_hints': type_equivalence_hints,
        # 'type_equivalence_hints_inverse' is the inverse of type_equivalence_hints,
        # which is always invertible.
        'type_equivalence_hints_inverse': invert_dict(type_equivalence_hints),
    }

    # Add the query root basic block to the output.
    basic_blocks = [
        blocks.QueryRoot({base_start_type})
    ]

    # Ensure the GraphQL query root doesn't immediately have a fragment (type coercion).
    # Instead of starting at one type and coercing to another,
    # users should simply start at the type to which they are coercing.
    immediate_fragment = _get_inline_fragment(base_ast)
    if immediate_fragment is not None:
        msg_args = {
            'coerce_to': immediate_fragment.type_condition.name.value,
            'type_from': base_start_type,
        }
        raise GraphQLCompilationError(u'Found inline fragment coercing to type {coerce_to}, '
                                      u'immediately inside query root asking for type {type_from}. '
                                      u'This is a contrived pattern -- you should simply start '
                                      u'your query at {coerce_to}.'.format(**msg_args))

    # Ensure the GraphQL query root doesn't have any vertex directives
    # that are disallowed on the root node.
    validate_root_vertex_directives(base_ast)

    # Compile and add the basic blocks for the query's base AST vertex.
    new_basic_blocks = _compile_ast_node_to_ir(
        schema, current_schema_type, base_ast, location, context)
    basic_blocks.extend(new_basic_blocks)

    # All operations after this point affect the global query scope, and are not related to
    # the "current" location in the query produced by the sequence of Traverse/Backtrack blocks.
    basic_blocks.append(blocks.GlobalOperationsStart())

    # Add any filters that apply to the global query scope.
    basic_blocks.extend(context['global_filters'])

    # Based on the outputs context data, add an output step and construct the output metadata.
    outputs_context = context['outputs']
    basic_blocks.append(_compile_output_step(outputs_context))
    output_metadata = {
        name: OutputMetadata(type=value['type'], optional=value['optional'])
        for name, value in six.iteritems(outputs_context)
    }

    return IrAndMetadata(
        ir_blocks=basic_blocks,
        input_metadata=context['inputs'],
        output_metadata=output_metadata,
        query_metadata_table=context['metadata'])


def _compile_output_step(outputs):
    """Construct the final ConstructResult basic block that defines the output format of the query.

    Args:
        outputs: dict, output name (string) -> output data dict, specifying the location
                 from where to get the data, and whether the data is optional (and therefore
                 may be missing); missing optional data is replaced with 'null'

    Returns:
        a ConstructResult basic block that constructs appropriate outputs for the query
    """
    if not outputs:
        raise GraphQLCompilationError(u'No fields were selected for output! Please mark at least '
                                      u'one field with the @output directive.')

    output_fields = {}
    for output_name, output_context in six.iteritems(outputs):
        location = output_context['location']
        optional = output_context['optional']
        graphql_type = output_context['type']

        expression = None
        existence_check = None
        # pylint: disable=redefined-variable-type
        if isinstance(location, FoldScopeLocation):
            if optional:
                raise AssertionError(u'Unreachable state reached, optional in fold: '
                                     u'{}'.format(output_context))

            if location.field == COUNT_META_FIELD_NAME:
                expression = expressions.FoldCountContextField(location)
            else:
                expression = expressions.FoldedContextField(location, graphql_type)
        else:
            expression = expressions.OutputContextField(location, graphql_type)

            if optional:
                existence_check = expressions.ContextFieldExistence(location.at_vertex())

        if existence_check:
            expression = expressions.TernaryConditional(
                existence_check, expression, expressions.NullLiteral)
        # pylint: enable=redefined-variable-type

        output_fields[output_name] = expression

    return blocks.ConstructResult(output_fields)


def _preprocess_graphql_string(graphql_string):
    """Apply any necessary preprocessing to the input GraphQL string, returning the new version."""
    # HACK(predrag): Workaround for graphql-core issue, to avoid needless errors:
    #                https://github.com/graphql-python/graphql-core/issues/98
    return graphql_string + '\n'


def _validate_schema_and_ast(schema, ast):
    """Validate the supplied graphql schema and ast.

    This method wraps around graphql-core's validation to enforce a stricter requirement of the
    schema -- all directives supported by the compiler must be declared by the schema, regardless of
    whether each directive is used in the query or not.

    Args:
        schema: GraphQL schema object, created using the GraphQL library
        ast: abstract syntax tree representation of a graphql query

    Returns:
        list containing schema and/or query validation errors
    """
    core_graphql_errors = validate(schema, ast)

    # The following directives appear in the core-graphql library, but are not supported by the
    # graphql compiler.
    unsupported_default_directives = frozenset([
        frozenset([
            'include',
            frozenset(['FIELD', 'FRAGMENT_SPREAD', 'INLINE_FRAGMENT']),
            frozenset(['if'])
        ]),
        frozenset([
            'skip',
            frozenset(['FIELD', 'FRAGMENT_SPREAD', 'INLINE_FRAGMENT']),
            frozenset(['if'])
        ]),
        frozenset([
            'deprecated',
            frozenset(['ENUM_VALUE', 'FIELD_DEFINITION']),
            frozenset(['reason'])
        ])
    ])

    # Directives expected by the graphql compiler.
    expected_directives = {
        frozenset([
            directive.name,
            frozenset(directive.locations),
            frozenset(six.viewkeys(directive.args))
        ])
        for directive in DIRECTIVES
    }

    # Directives provided in the parsed graphql schema.
    actual_directives = {
        frozenset([
            directive.name,
            frozenset(directive.locations),
            frozenset(six.viewkeys(directive.args))
        ])
        for directive in schema.get_directives()
    }

    # Directives missing from the actual directives provided.
    missing_directives = expected_directives - actual_directives
    if missing_directives:
        missing_message = (u'The following directives were missing from the '
                           u'provided schema: {}'.format(missing_directives))
        core_graphql_errors.append(missing_message)

    # Directives that are not specified by the core graphql library. Note that Graphql-core
    # automatically injects default directives into the schema, regardless of whether
    # the schema supports said directives. Hence, while the directives contained in
    # unsupported_default_directives are incompatible with the graphql-compiler, we allow them to
    # be present in the parsed schema string.
    extra_directives = actual_directives - expected_directives - unsupported_default_directives
    if extra_directives:
        extra_message = (u'The following directives were supplied in the given schema, but are not '
                         u'not supported by the GraphQL compiler: {}'.format(extra_directives))
        core_graphql_errors.append(extra_message)

    return core_graphql_errors

##############
# Public API #
##############


def graphql_to_ir(schema, graphql_string, type_equivalence_hints=None):
    """Convert the given GraphQL string into compiler IR, using the given schema object.

    Args:
        schema: GraphQL schema object, created using the GraphQL library
        graphql_string: string containing the GraphQL to compile to compiler IR
        type_equivalence_hints: optional dict of GraphQL interface or type -> GraphQL union.
                                Used as a workaround for GraphQL's lack of support for
                                inheritance across "types" (i.e. non-interfaces), as well as a
                                workaround for Gremlin's total lack of inheritance-awareness.
                                The key-value pairs in the dict specify that the "key" type
                                is equivalent to the "value" type, i.e. that the GraphQL type or
                                interface in the key is the most-derived common supertype
                                of every GraphQL type in the "value" GraphQL union.
                                Recursive expansion of type equivalence hints is not performed,
                                and only type-level correctness of this argument is enforced.
                                See README.md for more details on everything this parameter does.
                                *****
                                Be very careful with this option, as bad input here will
                                lead to incorrect output queries being generated.
                                *****

    Returns:
        IrAndMetadata named tuple, containing fields:
        - ir_blocks: a list of IR basic block objects
        - input_metadata: a dict of expected input parameters (string) -> inferred GraphQL type
        - output_metadata: a dict of output name (string) -> OutputMetadata object
        - query_metadata_table: a QueryMetadataTable object containing location metadata

    Raises flavors of GraphQLError in the following cases:
        - if the query is invalid GraphQL (GraphQLParsingError);
        - if the query doesn't match the schema (GraphQLValidationError);
        - if the query has more than one definition block (GraphQLValidationError);
        - if the query has more than one selection in the root object (GraphQLCompilationError);
        - if the query does not obey directive usage rules (GraphQLCompilationError);
        - if the query provides invalid / disallowed / wrong number of arguments
          for a directive (GraphQLCompilationError).

    In the case of implementation bugs, could also raise ValueError, TypeError, or AssertionError.
    """
    graphql_string = _preprocess_graphql_string(graphql_string)
    try:
        ast = parse(graphql_string)
    except GraphQLSyntaxError as e:
        raise GraphQLParsingError(e)

    validation_errors = _validate_schema_and_ast(schema, ast)

    if validation_errors:
        raise GraphQLValidationError(u'String does not validate: {}'.format(validation_errors))

    if len(ast.definitions) != 1:
        raise AssertionError(u'Unsupported graphql string with multiple definitions, should have '
                             u'been caught in validation: \n{}\n{}'.format(graphql_string, ast))
    base_ast = ast.definitions[0]

    return _compile_root_ast_to_ir(schema, base_ast, type_equivalence_hints=type_equivalence_hints)
